From 19c2918ad18ab8c22e8a2489d171a809788b190b Mon Sep 17 00:00:00 2001
From: Neal Garber <neal.garber@energyeast.com>
Date: Wed, 26 Aug 2009 10:21:32 -0400
Subject: [PATCH] Correct calculation of ntlm_auth challenge hash

---
 src/include/radius.h                               |    1 +
 .../types/rlm_eap_mschapv2/rlm_eap_mschapv2.c      |   22 +++++-
 src/modules/rlm_mschap/rlm_mschap.c                |   86 +++++++++++++++++++-
 3 files changed, 105 insertions(+), 4 deletions(-)

diff --git a/src/include/radius.h b/src/include/radius.h
index c7946a9..b5679d6 100644
--- a/src/include/radius.h
+++ b/src/include/radius.h
@@ -201,6 +201,7 @@
 #define PW_RESPONSE_PACKET_TYPE		1080
 #define PW_DIGEST_HA1			1081
 #define PW_MS_CHAP_USE_NTLM_AUTH	1082
+#define PW_NTLM_USER_NAME		1083
 #define PW_PACKET_SRC_IP_ADDRESS	1084
 #define PW_PACKET_DST_IP_ADDRESS	1085
 #define PW_PACKET_SRC_PORT		1086
diff --git a/src/modules/rlm_eap/types/rlm_eap_mschapv2/rlm_eap_mschapv2.c b/src/modules/rlm_eap/types/rlm_eap_mschapv2/rlm_eap_mschapv2.c
index 20e7875..9cf9f0d 100644
--- a/src/modules/rlm_eap/types/rlm_eap_mschapv2/rlm_eap_mschapv2.c
+++ b/src/modules/rlm_eap/types/rlm_eap_mschapv2/rlm_eap_mschapv2.c
@@ -360,7 +360,7 @@ static int mschapv2_authenticate(void *arg, EAP_HANDLER *handler)
 	int rcode;
 	mschapv2_opaque_t *data;
 	EAP_DS *eap_ds = handler->eap_ds;
-	VALUE_PAIR *challenge, *response;
+	VALUE_PAIR *challenge, *response, *name;
 
 	rad_assert(handler->request != NULL);
 	rad_assert(handler->stage == AUTHENTICATE);
@@ -503,11 +503,31 @@ static int mschapv2_authenticate(void *arg, EAP_HANDLER *handler)
 	response->vp_strvalue[1] = eap_ds->response->type.data[5 + MSCHAPV2_RESPONSE_LEN];
 
 	/*
+	 * Create NTLM-User-Name attribute from the MS-CHAP
+	 * response packet name field so rlm_mschap can use
+	 * it to create the challenge hash for ntlm_auth
+	 */
+	name = pairmake("NTLM-User-Name", "0x00", T_OP_EQ);
+	if (!name) {
+		pairfree(&challenge);
+		pairfree(&response);
+		radlog(L_ERR, "rlm_eap_mschapv2: Error creating NTLM-User-Name attribute: %s", fr_strerror());
+		return 0;
+	}
+
+	name->length = (((eap_ds->response->type.data[2] << 8) |
+			  eap_ds->response->type.data[3]) -
+			  eap_ds->response->type.data[4] - 5);
+	memcpy(name->vp_strvalue, &eap_ds->response->type.data[4 + MSCHAPV2_RESPONSE_LEN],
+	       name->length);
+
+	/*
 	 *	Add the pairs to the request, and call the 'mschap'
 	 *	module.
 	 */
 	pairadd(&handler->request->packet->vps, challenge);
 	pairadd(&handler->request->packet->vps, response);
+	pairadd(&handler->request->packet->vps, name);
 
 	/*
 	 *	If this options is set, then we do NOT authenticate the
diff --git a/src/modules/rlm_mschap/rlm_mschap.c b/src/modules/rlm_mschap/rlm_mschap.c
index d3159c3..eeefb81 100644
--- a/src/modules/rlm_mschap/rlm_mschap.c
+++ b/src/modules/rlm_mschap/rlm_mschap.c
@@ -285,7 +285,7 @@ static size_t mschap_xlat(void *instance, REQUEST *request,
 	uint8_t		*data = NULL;
 	uint8_t		buffer[32];
 	VALUE_PAIR	*user_name;
-	VALUE_PAIR	*chap_challenge, *response;
+	VALUE_PAIR	*chap_challenge, *response, *response_name;
 	rlm_mschap_t	*inst = instance;
 
 	response = NULL;
@@ -320,6 +320,7 @@ static size_t mschap_xlat(void *instance, REQUEST *request,
 			 */
 		} else if (chap_challenge->length == 16) {
 			char *username_string;
+			char *responsename_string;
 
 			RDEBUG2(" mschap2: %02x", chap_challenge->vp_octets[0]);
 			response = pairfind(request->packet->vps,
@@ -340,7 +341,7 @@ static size_t mschap_xlat(void *instance, REQUEST *request,
 			user_name = pairfind(request->packet->vps,
 					     PW_USER_NAME);
 			if (!user_name) {
-				RDEBUG2("User-Name is required to calculateMS-CHAPv1 Challenge.");
+				RDEBUG2("User-Name is required to calculate MS-CHAPv1 Challenge.");
 				return 0;
 			}
 
@@ -359,10 +360,49 @@ static size_t mschap_xlat(void *instance, REQUEST *request,
 			}
 
 			/*
+			 *      Check for NTLM-User-Name and if found, use it
+			 *      to construct the MSCHAPv1 challenge.  This is
+			 *      set by rlm_eap_mschap to the MS-CHAP Response
+			 *      packet Name field
+			 */
+			response_name = pairfind(request->packet->vps, PW_NTLM_USER_NAME);
+			if (!response_name) {
+				RDEBUG2("  No NTLM-User-Name attribute found.  Using User-Name to create MS-CHAPv1 challenge");
+			} else {
+				if ((responsename_string = strchr(response_name->vp_strvalue, '\\')) != NULL) {
+					if (inst->with_ntdomain_hack) {
+						responsename_string++;
+					} else {
+						RDEBUG2("  NT domain delimeter found in NTLM-User-Name, should we have enabled with_ntdomain_hack?");
+						responsename_string = response_name->vp_strvalue;
+					}
+				} else {
+					responsename_string = response_name->vp_strvalue;
+				}
+
+				RDEBUG2("  Comparing Response Name (%s) with User-Name (%s)", responsename_string, username_string);
+				if (strlen(username_string) != strlen(responsename_string) ||
+				    strncasecmp(username_string, responsename_string, strlen(username_string)) != 0) {
+					radlog_request(L_AUTH, 0, request, "User-Name (%s) is different than Response Name (%s)",
+						       username_string, responsename_string);
+					return RLM_MODULE_REJECT;
+				} else {
+					/*
+					 *      Response name is same as User-Name except perhaps for case
+					 *      Use Response name to construct MS-CHAPv1 challenge to avoid
+					 *      ntlm_auth Logon Failure
+					 */
+					RDEBUG2("  Using MS-CHAP Response Name (%s) to construct MS-CHAPv1 challenge", responsename_string);
+					username_string = responsename_string;
+				}
+			}
+
+			/*
 			 *	Get the MS-CHAPv1 challenge
 			 *	from the MS-CHAPv2 peer challenge,
 			 *	our challenge, and the user name.
 			 */
+			RDEBUG2("rlm_mschap: mschap_xlat: Creating challenge hash with username: %s", username_string);
 			challenge_hash(response->vp_octets + 2,
 				       chap_challenge->vp_octets,
 				       username_string, buffer);
@@ -1006,12 +1046,14 @@ static int mschap_authenticate(void * instance, REQUEST *request)
 #define inst ((rlm_mschap_t *)instance)
 	VALUE_PAIR *challenge = NULL;
 	VALUE_PAIR *response = NULL;
+	VALUE_PAIR *response_name = NULL;
 	VALUE_PAIR *password = NULL;
 	VALUE_PAIR *lm_password, *nt_password, *smb_ctrl;
 	VALUE_PAIR *username;
 	uint8_t nthashhash[16];
 	char msch2resp[42];
 	char *username_string;
+	char *responsename_string;
 	int chap = 0;
 	int		do_ntlm_auth;
 
@@ -1226,7 +1268,6 @@ static int mschap_authenticate(void * instance, REQUEST *request)
 			return RLM_MODULE_INVALID;
 		}
 
-
 		/*
 		 *	with_ntdomain_hack moved here
 		 */
@@ -1241,6 +1282,44 @@ static int mschap_authenticate(void * instance, REQUEST *request)
 		        username_string = username->vp_strvalue;
 		}
 
+		/*
+		 *      Check for NTLM-User-Name and if found, use it
+		 *      to construct the MSCHAPv1 challenge.  This is
+		 *      set by rlm_eap_mschap to the MS-CHAP Response
+		 *      packet Name field
+		 */
+		response_name = pairfind(request->packet->vps, PW_NTLM_USER_NAME);
+		if (!response_name) {
+			RDEBUG2("  No NTLM-User-Name attribute found.  Using User-Name to create MS-CHAPv1 challenge");
+		} else {
+			if ((responsename_string = strchr(response_name->vp_strvalue, '\\')) != NULL) {
+				if (inst->with_ntdomain_hack) {
+					responsename_string++;
+				} else {
+					RDEBUG2("  NT domain delimeter found in NTLM-User-Name, should we have enabled with_ntdomain_hack?");
+					responsename_string = response_name->vp_strvalue;
+				}
+			} else {
+				responsename_string = response_name->vp_strvalue;
+			}
+
+			RDEBUG2("  Comparing Response Name (%s) with User-Name (%s)", responsename_string, username_string);
+			if (strlen(username_string) != strlen(responsename_string) ||
+			    strncasecmp(username_string, responsename_string, strlen(username_string)) != 0) {
+				radlog_request(L_AUTH, 0, request, "User-Name (%s) is different than Response Name (%s)",
+					       username_string, responsename_string);
+				return RLM_MODULE_REJECT;
+			} else {
+				/*
+				 *      Response name is same as User-Name except perhaps for case
+				 *      Use Response name to construct MS-CHAPv1 challenge to avoid
+				 *      ntlm_auth Logon Failure
+				 */
+				RDEBUG2("  Using MS-CHAP Response Name (%s) to construct MS-CHAPv1 challenge", responsename_string);
+				username_string = responsename_string;
+			}
+		}
+
 #ifdef __APPLE__
 		/*
 		 *	No "known good" NT-Password attribute.  Try to do
@@ -1258,6 +1337,7 @@ static int mschap_authenticate(void * instance, REQUEST *request)
 		 *	MS-CHAPv2 takes some additional data to create an
 		 *	MS-CHAPv1 challenge, and then does MS-CHAPv1.
 		 */
+		RDEBUG2("rlm_mschap: mschap_authenticate: Creating challenge hash with username: %s", username_string);
 		challenge_hash(response->vp_octets + 2, /* peer challenge */
 			       challenge->vp_octets, /* our challenge */
 			       username_string,	/* user name */
-- 
1.6.4.1

